/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "private.h"

void copy_apr_to_node_entry(artnet_node_entry e, artnet_reply_t *reply) ;

		
/** 
 * Creates a new ArtNet node.
 * Takes a string containing the ip address to bind to, if the string is NULL 
 * it uses the first non loopback address
 *
 * @param ip the IP address to bind to
 * @return an artnet_node
 */
artnet_node artnet_init(const char *ip) {
	node n ;
	int ret, i ;
	
	n = Malloc(sizeof(artnet_node_t ) ) ;
	
	memset(n, 0x0, sizeof(artnet_node_t) ) ;

	// init node listing
	n->node_list.first = NULL ;
	n->node_list.current = NULL ;
	n->node_list.last = NULL ;
	n->node_list.length = 0;

	ret = net_init(n, ip) ;
	if (ret) return NULL ;

	// now setup the default parameters
	n->state.send_apr_on_change = FALSE ;
	n->state.reply_addr.s_addr = n->state.ip_addr.s_addr ;
	n->state.ar_count = 0 ;
	n->state.report_code = ARTNET_RCPOWEROK ;
	n->state.reply_addr.s_addr = 0 ;
	n->state.mode = ARTNET_STANDBY ;
	
	// set all ports to MERGE HTP mode and disable
	for(i=0; i < ARTNET_MAX_PORTS;i++) {
		n->ports.out[i].merge_mode = ARTNET_MERGE_HTP ;
		n->ports.out[i].enabled = FALSE ;
	}

	return (artnet_node) n;
}

/**
 * Starts the ArtNet node.
 * Binds to the network sockets and sends an ArtPoll 
 *
 * @param vn the artnet_node
 * @return 0 on success, -1 on failure
 *
 */
int artnet_start(artnet_node vn) {
	node n = (node) vn ;
	
	if (n == NULL)
		return -1 ;
	
	if( n->state.mode != ARTNET_STANDBY)
		return -1 ;

	printf("turning on\n" ) ;

	net_start(n) ;

	n->state.mode = ARTNET_ON ;

	if(n->state.reply_addr.s_addr == 0) {
		n->state.reply_addr = n->state.bcast_addr ;
	}

	// build the initial reply
	_build_art_poll_reply(n) ;
	
	// send a reply on startup
	send_poll_reply(n, FALSE) ;

	return 0 ;
}

/**
 * Stops the ArtNet node
 * Closes the network sockets held by the node
 *
 * @param vn the artnet_node
 * @return 0 on success, -1 on failure
 */
int artnet_stop(artnet_node vn) {
	node n = (node) vn ;
	
	if ( n == NULL) 
		return -1;

	if(n->state.mode != ARTNET_ON) 
		return -1 ;

	net_close(n) ;
	n->state.mode = ARTNET_STANDBY ;

	printf("turning off\n") ;
	
	return 0; 
}

/**
 * Handle any received packets.
 * This function is the workhorse of libartnet. You have a couple of options:
 * 	- use artnet_get_sd() to retrieve the socket descriptors and select to
 * 		detect network activity. Then call artnet_read(node,0) 
 * 		when activity is detected.
 * 	- call artnet_read repeatedaly from within a loop with an appropriate 
 * 		timeout
 *
 * @param vn the artnet_node
 * @param timeout the number of seconds to block for if nothing is pending
 * @return 0 on success, -1 on failure
 */
int artnet_read(artnet_node vn, int timeout) {
	node n = (node) vn ;	
	artnet_packet_t p ;
	int done = FALSE ;
	
	if ( n == NULL) 
		return -1;

	if (n->state.mode != ARTNET_ON) 
		return -1 ;

	while(! done) {
		memset( &p.data, 0x0, sizeof(p.data) ) ;
			
		p.length = net_recv(n, &p, timeout) ;

		if( p.length > MIN_PACKET_SIZE && get_type(&p) )
			handle(n, &p) ;
		else if (p.length == 0)
			// no more data
			done = TRUE ;
	} while (! done) ;
	
	return 0 ;
}

/**
 * This is used to set handlers for received artnet packets.
 *
 * @param vn 		The artnet_node
 * @param handler	The handler to set
 * @param fh		A pointer to a function, set to NULL to turn off
 * @param data		Data to be passed to the handler when its called
 *
 * @return 0 on sucess, -1 on failure
 *
 */
int artnet_set_handler(artnet_node vn, artnet_handler_name_t handler, int (*fh)(artnet_node n, void *pp, void * d), void *data) {
	node n = (node) vn ;
	callback_t *callback ;

	if ( n == NULL) 
		return -1;

	switch( handler) {
		case ARTNET_RECV_HANDLER:
			callback = &n->callbacks.recv ;
			break;
		case ARTNET_SEND_HANDLER:
			callback = &n->callbacks.send ;
			break;
		case ARTNET_POLL_HANDLER:
			callback = &n->callbacks.poll ;
			break ;
		case ARTNET_REPLY_HANDLER:
			callback = &n->callbacks.reply ;
			break ;
		case ARTNET_ADDRESS_HANDLER:
			callback = &n->callbacks.address ;
			break ;
		case ARTNET_DMX_HANDLER:
			callback = &n->callbacks.dmx ;
			break ;
		default:
			err_warn("Invalid handler defined") ;
			return 1;
			break;
	}
	callback->fh = fh ;
	callback->data = data;
	
	return 0;

}



int artnet_set_dmx_handler(artnet_node vn, int (*fh)(artnet_node n,int port, void *d), void *data ) {
	node n = (node) vn ;
	callback_t *callback ;

	if ( n == NULL) 
		return -1;
	
	n->callbacks.dmx_c.fh = fh ;
	n->callbacks.dmx_c.data = data;

	return 0;
}


// sends a poll to the specified ip, or if null, will broadcast
// talk_to_me - modify remote nodes behaviour, see spec
// FIX ME - this should clear the node list - but this will cause issues if the caller holds references
//   to certain nodes
//

/**
 *
 * @param vn the artnet_node
 * @param ip the ip address to send to, NULL will broadcast the ArtPoll
 * @param talk_to_me the value for the talk to me
 */
int artnet_send_poll(artnet_node vn, const char *ip,  artnet_ttm_value_t talk_to_me) {
	node n = (node) vn ;
	artnet_packet_t p ;

	if ( n == NULL) 
		return -1;

	if (n->state.mode != ARTNET_ON) 
		return -1 ;

	
	if(n->state.node_type == ARTNET_SRV || n->state.node_type == ARTNET_RAW) {
		// check if caller specified an ip or wanted bcase
		if( ip != NULL) {
			if( ! inet_aton(ip, &p.to) ) {
				err_warn("ip conversion from %s failed", ip) ;
				return -1 ;
			}
		} else {
			p.to.s_addr = n->state.bcast_addr.s_addr ;
		}
	
		memcpy( &p.data.ap.id, ARTNET_STRING, ARTNET_STRING_SIZE) ;
		p.data.ap.opCode =  ARTNET_POLL;
		p.data.ap.verH = 0;
		p.data.ap.ver = ARTNET_VERSION;
		p.data.ap.ttm = ~talk_to_me ;
		p.data.ap.pad = 0 ;
		
		p.length = sizeof(artnet_poll_t) ;
		net_send(n, &p) ;


		printf("Sending Poll\n") ;
	} else {
		err_warn("Not sending poll, not a server or raw device") ;
		return -1 ;
	}
	return 0 ;


}



int artnet_send_poll_reply(artnet_node vn) {
	node n = (node) vn ;

	if ( n == NULL) 
		return -1;

	
	if (n->state.mode != ARTNET_ON) 
		return -1 ;

	send_poll_reply(n, FALSE) ;

	return 0 ;
}



int artnet_send_dmx(artnet_node vn, int port_id, int16_t length, uint8_t *data ) {
	node n = (node) vn ;
	artnet_packet_t p ;

	if ( n == NULL) 
		return -1;

	if (port_id < 0 || port_id >= ARTNET_MAX_PORTS) 
		return -1;

	if( length < 2 || length > 512) 
		err_warn("Length of dmx data out of bounds (should be between 2 and 512, was %i)\n", length) ;
	
	if (n->state.mode != ARTNET_ON) 
		return -1 ;

	if(n->ports.in[port_id].port_status & PORT_STATUS_DISABLED_MASK)
		return -1 ;
	
	// ok we're going to send now, make sure we turn the activity bit on
	n->ports.in[port_id].port_status = n->ports.in[port_id].port_status | PORT_STATUS_ACT_MASK ;
	
	// set dst addr and length
	p.to.s_addr = n->state.bcast_addr.s_addr ;

	p.length = sizeof(artnet_dmx_t) - (ARTNET_MAX_DMX - length);
	
	// now build packet
	memcpy( &p.data.admx.id, ARTNET_STRING, ARTNET_STRING_SIZE) ;
	p.data.admx.opCode =  ARTNET_DMX;
	p.data.admx.verH = 0;
	p.data.admx.ver = ARTNET_VERSION;
	p.data.admx.sequence = n->ports.in[port_id].seq ;
	p.data.admx.physical = port_id ;
	p.data.admx.universe = n->ports.in[port_id].port_addr ;
	
	// set length
	p.data.admx.lengthHi = short_get_high_byte(length) ; 
	p.data.admx.length = short_get_low_byte(length) ;
	memcpy(&p.data.admx.data, data, length) ; 

	net_send(n, &p) ;
	
	n->ports.in[port_id].seq++ ;

	return 0 ;
}



int artnet_send_address(artnet_node vn, artnet_node_entry e,const char *shortName, const char *longName, uint8_t inAddr[ARTNET_MAX_PORTS], uint8_t outAddr[ARTNET_MAX_PORTS], uint8_t subAddr, artnet_port_command_t cmd ) {
	node n = (node) vn ;
	artnet_node_entry_t *entry = (artnet_node_entry_t *) e ;
	
	artnet_packet_t p ;

	if ( n == NULL || e == NULL ) 
		return -1;

	if(n->state.node_type == ARTNET_SRV || n->state.node_type == ARTNET_RAW) {
		// set dst, type and length
		p.to.s_addr = bytes_to_32(e->ip) ;

		p.length = sizeof(artnet_address_t) ;
		p.type = ARTNET_ADDRESS ;
		
		// now build packet, copy the number of ports from the reply recieved from this node
		memcpy( &p.data.addr.id, ARTNET_STRING, ARTNET_STRING_SIZE) ;
		p.data.addr.opCode =  ARTNET_ADDRESS;
		p.data.addr.verH = 0;
		p.data.addr.ver = ARTNET_VERSION;
		p.data.addr.filler1 = 0;
		p.data.addr.filler2 = 0;
		strncpy((char*) &p.data.addr.shortname, shortName, ARTNET_SHORT_NAME_LENGTH) ;
		strncpy((char*) &p.data.addr.longname, longName, ARTNET_LONG_NAME_LENGTH) ;

		memcpy(&p.data.addr.swin, inAddr, ARTNET_MAX_PORTS) ;
		memcpy(&p.data.addr.swout, outAddr, ARTNET_MAX_PORTS) ;

		p.data.addr.subnet = subAddr;
		p.data.addr.swvideo = 0x00;
		p.data.addr.command = cmd ;
		net_send(n, &p) ;
	}

	return 0 ;

}


/**
 *
 * Sends an ArtInput packet to the specified node, this packet is used to 
 * enable/disable the input ports on the remote node.
 *
 * 0x01 disable port
 * 0x00 enable port
 *
 * NOTE: should have enums here instead of uint8_t for settings
 *
 */
int artnet_send_input(artnet_node vn, artnet_node_entry e, uint8_t settings[ARTNET_MAX_PORTS]) {
	node n = (node) vn ;
	artnet_node_entry_t *entry = (artnet_node_entry_t *) e ;
	
	artnet_packet_t p ;

	if ( n == NULL || e == NULL ) 
		return -1;

	if(n->state.node_type == ARTNET_SRV || n->state.node_type == ARTNET_RAW) {
		// set dst, type and length
		p.to.s_addr = bytes_to_32(e->ip) ;
		p.length = sizeof(artnet_input_t) ;
		p.type = ARTNET_INPUT ;
		
		// now build packet, copy the number of ports from the reply recieved from this node
		memcpy( &p.data.ainput.id, ARTNET_STRING, ARTNET_STRING_SIZE) ;
		p.data.ainput.opCode =  ARTNET_INPUT;
		p.data.ainput.verH = 0;
		p.data.ainput.ver = ARTNET_VERSION;
		p.data.ainput.filler1 = 0;
		p.data.ainput.filler2 = 0;
		p.data.ainput.numbportsH = short_get_high_byte(e->numbports) ;
		p.data.ainput.numbports = short_get_low_byte(e->numbports) ;
		memcpy(&p.data.ainput.input, &settings, ARTNET_MAX_PORTS) ;
		net_send(n, &p) ;
	}

	return 0 ;
}

// need address here
int artnet_send_ipprog_reply(artnet_node vn) {
	node n = (node) vn ;
	artnet_packet_t ipr ;

	//memset(ipr.data, 0x00, sizeof(artnet_ipprog_reply_t) );
	

	



//	net_send(n, &reply) ; 
	return 0 ;
}



uint8_t *artnet_read_dmx(artnet_node vn, int port_id, int *length) {
	node n = (node) vn ;

	if ( n == NULL) 
		return NULL;

	if (port_id < 0 || port_id >= ARTNET_MAX_PORTS) 
		return NULL;

	*length = n->ports.out[port_id].length ;
	
	return &n->ports.out[port_id].data[0] ;

}

//--------------------------------------
// Functions to change the node state (setters)

// type : server, node, mserver, raw
int artnet_set_node_type(artnet_node vn, artnet_node_type type) {
	node n = (node) vn ;

	if ( n == NULL) 
		return -1;

	n->state.node_type = type ;
	return 0;
}


/**
 * Sets the artnet subnet address for this node.
 * The subnet address has nothing to do with IP addresses). An ArtNet subnet is a grouping of 16 DMX universes
 * (ie. ports)
 * 
 * The subnet address is between 0 and 15. If the supplied address is larger than 15, the
 * lower 4 bits will be used in setting the address.
 *
 * It will have no effect if the node is under network control.
 *
 * Note that changing the subnet address will cause the universe addresses of all ports to change.
 *
 * @param vn the artnet_node
 * @param subnet new subnet address
 *
 */
int artnet_set_subnet_addr(artnet_node vn, uint8_t subnet) {
	node n = (node) vn ;
	int i ;
	if ( n == NULL) 
		return -1;

	n->state.default_subnet = subnet ;
 
	// if not under network control, and the subnet is different from the current one
	if( ! n->state.subnet_net_ctl  && subnet != n->state.subnet) {
		n->state.subnet = subnet ;

		// redo the addresses for each port
		for(i =0 ; i < ARTNET_MAX_PORTS; i++ ) {
			n->ports.in[i].port_addr = ((n->state.subnet & LOW_NIBBLE) << 4) | (n->ports.in[i].port_addr & LOW_NIBBLE) ;
			// reset dmx sequence number
			n->ports.in[i].seq = 0 ;
			
			n->ports.out[i].port_addr = ((n->state.subnet & LOW_NIBBLE) << 4) | (n->ports.out[i].port_addr & LOW_NIBBLE) ;
		}

		_build_art_poll_reply(n) ;
		send_poll_reply(n,FALSE) ;
	} else if (n->state.subnet_net_ctl ) {
		//  trying to change subnet addr while under network control
		n->state.report_code = ARTNET_RCUSERFAIL ;
	}

	return 0 ;
}


/**
 * Sets the short name of the node. 
 * The string should be null terminated and a maxmium of 18 Characters will be used
 * 
 * @param vn the artnet_node
 * @param name the short name of the node.
 */
int artnet_set_short_name(artnet_node vn, const char *name) {
	node n = (node) vn ;
	if ( n == NULL) 
		return -1;

	strncpy((char *) &n->state.short_name, name, 18) ;
	n->state.short_name[17] = 0x00 ;
	_build_art_poll_reply(n) ;
	return 0;
	
}

/**
 * Sets the long name of the node. 
 * The string should be null terminated and a maximium of 64 characters will be used
 *
 * @param vn the artnet_node
 * @param name the node's long name
 */
int artnet_set_long_name(artnet_node vn, const char *name) {
	node n = (node) vn ;

	if ( n == NULL) 
		return -1;

	strncpy((char *) &n->state.long_name, name, 64) ;
	n->state.long_name[63] = 0x00 ;
	_build_art_poll_reply(n) ;
	return 0 ;
}

/*
 * Sets the direction and type of port
 *
 * 
 * @param vn the artnet_node
 * @param id
 * @param direction
 * @param data
 */
int artnet_set_port_type(artnet_node vn, int id, artnet_port_settings_t settings, artnet_port_data_t data ) {
	node n = (node) vn ;

	if ( n == NULL) 
		return -1;

	if (id < 0 || id >= ARTNET_MAX_PORTS) 
		return -1;

	n->ports.types[id] = settings | data ;

	return 0 ;
}


/**
 * Sets the port address of the port.
 *
 * Just to set some terms straight:
 *	- subnet address, is 4 bits, set on a per-node basis
 *	- port address, 4 bits, set on a per-port basis
 *	- universe address, 8 bits derrived from the subnet and port addresses, specific (but may
 *	not be unique) to a port.
 *
 * The upper four bits of the universe address are from the subnet address, while the lower 
 * four are from the port address.
 * 
 * So for example, if the subnet address of the node is 0x03, and the port address is
 * 0x02, the universe address for the port will be 0x32.
 *
 * As the port address is between 0 and 15, only the lower 4 bits of the addr argument 
 * will be used.
 *
 * The operation may have no affect if the port is under network control. 
 *
 * @param vn the artnet_node
 * @param id the phyiscal port number (from 0 to ARTNET_MAX_PORTS-1 )
 * @param dir either ARTNET_INPUT_PORT or ARTNET_OUTPUT_PORT
 * @param addr the new port address
 */
int artnet_set_port_addr(artnet_node vn, int id, artnet_port_dir_t dir, uint8_t addr) {
	node n = (node) vn ;

	port_t *port ;
	
	if(n == NULL)
		return -1;

	if (id < 0 || id >= ARTNET_MAX_PORTS) 
		return -1;

	if(addr > 16) {
		err_warn("Attempt to set port %i to invaild address %#hhx\n", id, addr) ;
		return -1 ;
	}

	if(dir == ARTNET_INPUT_PORT) {
		port = &n->ports.in[id].port ;
	} else if( dir == ARTNET_OUTPUT_PORT) {
		port = &n->ports.out[id].port ;
		n->ports.out[id].enabled = TRUE ;
	} else {
		err_warn("Invalid port direction\n") ;
		return -1;
	}

	port->default_addr = addr ;
	
	// if not under network control and address is changing
	if(! port->net_ctl && (addr & LOW_NIBBLE) != (port->addr & LOW_NIBBLE)) {
		port->addr = ((n->state.subnet & LOW_NIBBLE) << 4) | (addr & LOW_NIBBLE) ;

		// reset seq if input port
		if(dir == ARTNET_INPUT_PORT) 
			n->ports.in[id].seq = 0 ;

		_build_art_poll_reply(n) ;
		send_poll_reply(n,FALSE) ;
	} else if (port->net_ctl ) {
		//  trying to change port addr while under network control
		n->state.report_code = ARTNET_RCUSERFAIL ;
	}

	return 0;
}

/**
 * Dumps the node config to stdout.
 *
 * @param vn the artnet_node
 */
int artnet_dump_config(artnet_node vn) {
	node n = (node) vn ;

	if ( n == NULL) 
		return -1;
	
	printf("#### NODE CONFIG ####\n") ;
	printf("Node Type: %i\n", n->state.node_type) ;
	printf("Short Name: %s\n", n->state.short_name) ;
	printf("Long Name: %s\n", n->state.long_name) ;
	printf("Subnet: %#hx\n", n->state.subnet) ;
	printf("Default Subnet: %#hx\n", n->state.default_subnet) ;
	printf("Net Ctl: %i\n", n->state.subnet_net_ctl) ;
	printf("#####################\n") ;

	return 0;
}

/**
 * Returns the socket descriptor associated with this artnet_node.
 * libartnet currently uses two descriptors per node, one bound
 * to the network address and one bound to the subnet broadcast address
 *
 * @param vn the artnet_node
 * @param socket the index of the socket descriptor to fetch (0 or 1)
 * @return the socket descriptor
 */
int artnet_get_sd(artnet_node vn, int socket) {
	node n = (node) vn ;
	if ( n == NULL) 
		return -1;

	if(socket < 0 || socket > 1) 
		return -1 ;
	
	if( n->state.mode != ARTNET_ON)
		return -1 ;

	return n->sd[socket] ;
}

/**
 * Returns the artnet_node_list.
 * The artnet_node_list holds artnet_node_entry(s) that represent the discovered 
 * remote nodes on the network
 * NOTE: this function is not THREAD SAFE
 *
 * @param vn the artnet_node
 * @return the artnet_node_list
 */

// think about threading issues here
// crap, we might need a separate node list for each call ?

artnet_node_list artnet_get_nl(artnet_node vn) {
	node n = (node) vn ;

	if ( n == NULL) 
		return NULL;

	return &n->node_list ;

}

/**
 * Repositions the pointer to the first entry in the artnet_node_list
 * NOTE: this function is not THREAD SAFE
 * 
 * @param vnl the artnet_node_list
 * @return the first artnet_node_entry in the list, or NULL if the list is empty
 */
artnet_node_entry artnet_nl_first(artnet_node_list vnl) {
	node_list_t *nl = (node_list_t*) vnl ;

	if(nl == NULL)
		return NULL ;

	nl->current = nl->first ;

	return nl->current ;
	
	
}

/**
 * Moves the pointer to the next element in the artnet_node_list
 * NOTE: this function is not THREAD SAFE
 *
 * @param vnl the artnet_node_list
 * @return the next artnet_node_entry, or NULL if the end of the list is reached
 */
artnet_node_entry artnet_nl_next(artnet_node_list vnl) {
	node_list_t *nl = (node_list_t*) vnl ;

	if(nl == NULL)
		return NULL ;

	nl->current = nl->current->next ;
	return nl->current ;
}

/**
 * Returns the length of the artnet_node_list
 * NOTE: this function is not THREAD SAFE
 * 
 * @param vnl the artnet_node_list
 * @return the length of the list
 */
int artnet_nl_get_length(artnet_node_list vnl) {
	node_list_t *nl = (node_list_t*) vnl ;

	if(nl == NULL)
		return 0 ;

	return nl->length ;

}


void artnet_nl_update(node_list_t *nl, artnet_reply_t *reply) {
	artnet_node_entry new_entry ;
	int found = FALSE ;
	
	artnet_node_entry tmp = nl->first ;
	// check if this packet is in list
	while(tmp != NULL) {
		// byte ordering will affect this, fix me
		if( !memcmp(&reply->ip, &tmp->ip, 4) ) {
			found = TRUE;
			break ;
		}
		tmp =  tmp->next ;
	}

	if(!found) {
		// add to list	
		new_entry = (artnet_node_entry) Malloc(sizeof(artnet_node_entry_t)) ;
		copy_apr_to_node_entry(new_entry, reply) ;
		new_entry->next = NULL ;
		
		if(nl->first == NULL) {
			nl->first = new_entry ;
			nl->last = new_entry ;
		} else {
			nl->last->next = new_entry ;
			nl->last = new_entry;
		}
		nl->length++;
	} else {
		// update entry
		copy_apr_to_node_entry(tmp, reply) ;
	}

}


//-----------------------------------------------------------------------------
// Private functions follow
//-----------------------------------------------------------------------------


// this is called when the node's state changes to rebuild the 
// artpollreply packet
void _build_art_poll_reply(node n) {
	int i ;

	// shorten the amount we have to type
	artnet_reply_t *ar = &n->ar_temp ;

	memset(ar, 0x00, sizeof(artnet_reply_t) );
		
	memcpy( &ar->id, ARTNET_STRING, ARTNET_STRING_SIZE) ;
	ar->opCode =  ARTNET_REPLY;
	memcpy(&ar->ip , &n->state.ip_addr.s_addr, 4) ;
	ar->port = ARTNET_PORT ;
	ar->verH = 0;
	ar->ver = 0;
	ar->subH = 0;
	ar->sub = n->state.subnet ;
	ar->oemH = 0 ;
	ar->oem = OEM_CODE ;
	ar->ubea = 0 ;
	// ar->status
	
//	if(n->state

	// status need to be recalc everytime
	//ar->status 
	// what can we used for this ?
	ar->etsaman[0] = 's';
	ar->etsaman[1] = 'n';
		
	memcpy(&ar->shortname, &n->state.short_name, sizeof(n->state.short_name) ) ;	
	memcpy(&ar->longname, &n->state.long_name, sizeof(n->state.long_name) ) ;
	// the report is generated on every send

	// port stuff here
	ar->numbportsH = 0;
	ar->numbports = 4;

	for (i=0; i< ARTNET_MAX_PORTS; i++) {
		ar->porttypes[i] = n->ports.types[i] ;
		ar->goodinput[i] = n->ports.in[i].port_status ;
		ar->goodoutput[i] = n->ports.out[i].port_status ;
		ar->swin[i] = n->ports.in[i].port_addr ;
		ar->swout[i] = n->ports.out[i].port_addr ;
	}

	ar->swvideo  = 0;
	ar->swmacro = 0;
	ar->swremote = 0 ;
		
	// spares
	ar->sp1 =  0;
	ar->sp2 =  0;
	ar->sp3 =  0;

	// set style
	switch (n->state.node_type) {
		case ARTNET_SRV:
			ar->style = STSERVER ;
			break ;
		case ARTNET_MSRV:
			ar->style = STMEDIA ;
			break ;
		case ARTNET_NODE:
			ar->style = STNODE ;
			break;
		default:
			err_warn("node type not recognised!") ;
	}

	memcpy(	&ar->mac, &n->state.hw_addr, 6)  ;
//	ar->filler 			// set to 0x00 by memset above

}


void copy_apr_to_node_entry(artnet_node_entry e, artnet_reply_t *reply) {

	memcpy(&e->ip, &reply->ip, 4);
	e->ver = bytes_to_short(reply->verH, reply->ver);
	e->sub = bytes_to_short(reply->subH, reply->sub);
	e->oem = bytes_to_short(reply->oemH, reply->oem);
	e->ubea = reply->ubea ;
	memcpy(&e->etsaman, &reply->etsaman, 2) ;
	memcpy(&e->shortname, &reply->shortname,  sizeof(e->shortname));
	memcpy(&e->longname, &reply->shortname, sizeof(e->longname));
	memcpy(&e->nodereport, &reply->nodereport, sizeof(e->nodereport));
	e->numbports = bytes_to_short(reply->numbportsH, reply->numbports);
	memcpy(&e->porttypes, &reply->porttypes, ARTNET_MAX_PORTS) ;
	memcpy(&e->goodinput, &reply->goodinput, ARTNET_MAX_PORTS) ;
	memcpy(&e->goodinput, &reply->goodinput, ARTNET_MAX_PORTS) ;
	memcpy(&e->goodoutput, &reply->goodoutput, ARTNET_MAX_PORTS) ;
	memcpy(&e->swin, &reply->swin, ARTNET_MAX_PORTS) ;
	memcpy(&e->swout, &reply->swout, ARTNET_MAX_PORTS) ;
	e->swvideo = reply->swvideo ;
	e->swmacro = reply->swmacro ;
	e->swremote = reply->swremote ;
	e->style = reply->style ;
	memcpy(&e->mac, &reply->mac, 8);
}
