/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "private.h"

uint8_t _make_addr(uint8_t subnet, uint8_t addr) ;
void check_timeouts(node n, int port) ;
void merge(node n, int port, int length) ;

/** checks if the callback is defined, if so call it passing the packet and 
 * the user supplied data. 
 * If the callbacks return a non-zero result, further processing is canceled.
 */
int check_callback(node n, artnet_packet p, callback_t callback) {
	int ret ;
	if(callback.fh != NULL) 
		return callback.fh(n,p, callback.data) ;

	return 0;

}

/**
 * Handle an artpoll packet
 */
void handle_poll(node n, artnet_packet p) {

	// run callback if defined
	if(check_callback(n, p, n->callbacks.poll) )
		return ;

	if( n->state.node_type != ARTNET_RAW) {

		//if we're told to unicast further replies
		if(p->data.ap.ttm & TTM_REPLY_MASK) {
			n->state.reply_addr = p->from;
		} else {
			n->state.reply_addr.s_addr = n->state.bcast_addr.s_addr ;
		}

		// if we are told to send updates when node conditions change
		if(p->data.ap.ttm & TTM_BEHAVIOUR_MASK) {
			n->state.send_apr_on_change = TRUE ;
		} else {
			n->state.send_apr_on_change = FALSE ;
		}
	
		send_poll_reply(n, TRUE) ;

	} 
}

/**
 * handle an art poll reply
 */
void handle_reply(node n, artnet_packet p) {

	// update the node list
	artnet_nl_update(&n->node_list, p) ;
	
	// run callback if defined
	if(check_callback(n, p, n->callbacks.reply) )
		return ;

}

/**
 * handle a art dmx packet
 *
 *
 */
void handle_dmx(node n, artnet_packet p) {
	int i,j, data_length ;
	output_port_t *port ;
	in_addr_t ipA, ipB ;

	// run callback if defined	
	if(check_callback(n, p, n->callbacks.dmx) )
		return ;

	data_length = (int) bytes_to_short(p->data.admx.lengthHi, p->data.admx.length) ;
	data_length = min(data_length, ARTNET_MAX_DMX) ;

	// find matching output ports
	for(i=0; i < ARTNET_MAX_PORTS ; i++) {
//		printf("port %i %i == %i , %i \n", i, p->data.admx.universe , n->ports.out[i].port_addr ,n->ports.out[i].enabled) ;
		
		// if the addr matches and this port is enabled
		if (p->data.admx.universe == n->ports.out[i].port_addr &&
			n->ports.out[i].enabled == TRUE ) {

			port = &n->ports.out[i] ;
			ipA = port->ipA.s_addr ;
			ipB = port->ipB.s_addr ;

			// ok packet matches this port
			n->ports.out[i].port_status = n->ports.out[i].port_status | PORT_STATUS_ACT_MASK ;

			/**
			 * 9 cases for merging depending on what the stored ips are.
			 * here's the truth table
			 * 
			 *
			 * \   ipA   #           #            #             #
			 *  ------   #   empty   #            #             #
			 *   ipB  \  #   ( 0 )   #    p.from  #   ! p.from  # 
			 * ##################################################
			 *           # new node  # continued  # start       #
			 *  empty    # first     #  trans-    #  merge      #
			 *   (0)     #   packet  #   mission  #             #
			 * ##################################################
			 *           #continued  #            # cont        #
			 *  p.from   # trans-    # invalid!   #  merge      #
			 *           #  mission  #            #             #
			 * ##################################################
			 *           # start     # cont       #             #
			 * ! p.from  #  merge    #   merge    # discard     #
			 *           #           #            #             #
			 * ##################################################
			 *
			 * The merge exits when:
			 *   o ACCancel command is received in an ArtAddress packet
			 *       (this is done in handle_address )
			 *   o no data is recv'ed from one source in 10 seconds
			 *
			 */

			check_timeouts(n,i) ;
			
			if(ipA == 0 && ipB == 0) {
				// first packet recv on this port
				port->ipA.s_addr = p->from.s_addr ;
				port->timeA = time(NULL) ;
				
				memcpy(&port->dataA, &p->data.admx.data, data_length) ;
				port->length = data_length ;
				memcpy(&port->data, &p->data.admx.data, data_length) ;
			} 
			else if (ipA == p->from.s_addr && ipB == 0) {
				//continued transmission from the same ip (source A)

				port->timeA = time(NULL) ;
				memcpy(&port->dataA, &p->data.admx.data, data_length) ;
				port->length = data_length ;
				memcpy(&port->data, &p->data.admx.data, data_length) ;
			}
			else if (ipA == 0 && ipB == p->from.s_addr) {
				//continued transmission from the same ip (source B)

				port->timeB = time(NULL) ;
				memcpy(&port->dataB, &p->data.admx.data, data_length) ;
				port->length = data_length ;
				memcpy(&port->data, &p->data.admx.data, data_length) ;
			}
			else if (ipA != p->from.s_addr  && ipB == 0) {
				// new source, start the merge
				port->ipB.s_addr = p->from.s_addr ;
				port->timeB = time(NULL) ;
				memcpy(&port->dataB, &p->data.admx.data,data_length) ;
				port->length = data_length ;

				merge(n,i,data_length) ;

				// send reply if needed

			}
			else if (ipA == 0 && ipB == p->from.s_addr) {
				// new source, start the merge
				port->ipA.s_addr = p->from.s_addr ;
				port->timeB = time(NULL) ;
				memcpy(&port->dataB, &p->data.admx.data,data_length) ;
				port->length = data_length ;

				merge(n,i,data_length) ;
				
				// send reply if needed
			}
			else if (ipA == p->from.s_addr && ipB != p->from.s_addr) {
				// continue merge
				port->timeA = time(NULL) ;
				memcpy(&port->dataA, &p->data.admx.data,data_length) ;
				port->length = data_length ;

				// do merge
				merge(n,i,data_length) ;

			}
			else if (ipA != p->from.s_addr && ipB == p->from.s_addr) {
				// continue merge
				port->timeB = time(NULL) ;
				memcpy(&port->dataB, &p->data.admx.data,data_length) ;
				port->length = data_length ;

				// do merge
				merge(n,i,data_length) ;

			}
			else if (ipA == p->from.s_addr && ipB == p->from.s_addr) {
				err_warn("In handle_dmx, source matches both buffers, this shouldn't be happening!\n") ;

			} 
			else if (ipA != p->from.s_addr && ipB != p->from.s_addr) {
				err_warn("In handle_dmx, more than two sources, discarding data\n") ;

			} 
			else {
				err_warn("In handle_dmx, no cases matched, this shouldn't happen!\n") ;
				
			}

			// do the dmx callback here
			if(n->callbacks.dmx_c.fh != NULL)
				n->callbacks.dmx_c.fh(n,i, n->callbacks.dmx_c.data) ;
		}
	}
	return ;
}


/**
 * handle art address packet.
 * This can reprogram certain nodes settings such as short/long name, port
 * addresses, subnet address etc.
 *
 */
void handle_address(node n, artnet_packet p) {
	int i, old_subnet;
	int addr[ARTNET_MAX_PORTS] ;

	if(check_callback(n, p, n->callbacks.address) )
		return ;

	// servers (and raw nodes) don't respond to address packets
	if(n->state.node_type == ARTNET_SRV || n->state.node_type == ARTNET_RAW ) 
		return ;

	// reprogram shortname if required
	if(p->data.addr.shortname[0] != PROGRAM_DEFAULTS &&
	   p->data.addr.shortname[0] != PROGRAM_NO_CHANGE	) {
		memcpy(&n->state.short_name, &p->data.addr.shortname, ARTNET_SHORT_NAME_LENGTH) ;
		n->state.report_code = ARTNET_RCSHNAMEOK ;
	}
	// reprogram long name if required
	if(p->data.addr.longname[0] != PROGRAM_DEFAULTS &&
	   p->data.addr.longname[0] != PROGRAM_NO_CHANGE    ) {
		memcpy(&n->state.long_name, &p->data.addr.longname, ARTNET_LONG_NAME_LENGTH) ;
		n->state.report_code = ARTNET_RCLONAMEOK ;
	}
	
	// first of all store existing port addresses
	// then we can work out if they change
	for(i=0; i< ARTNET_MAX_PORTS; i++) {
		addr[i] = n->ports.in[i].port_addr  ;
	}

	// program subnet
	old_subnet = p->data.addr.subnet ;
	if (p->data.addr.subnet == PROGRAM_DEFAULTS)  {
		// reset to defaults
		n->state.subnet = n->state.default_subnet ;
		n->state.subnet_net_ctl = FALSE ;

	} else if( p->data.addr.swout[i] & PROGRAM_CHANGE_MASK) {
		n->state.subnet = p->data.addr.subnet & ~PROGRAM_CHANGE_MASK ;
		n->state.subnet_net_ctl = TRUE;

	}

	// check if subnet has actually changed
	if(old_subnet != n->state.subnet) {
		// if it does we need to change all port addresses
		for(i=0; i< ARTNET_MAX_PORTS; i++) {
			n->ports.in[i].port_addr = _make_addr(n->state.subnet, n->ports.in[i].port_addr) ;
			n->ports.out[i].port_addr = _make_addr(n->state.subnet, n->ports.out[i].port_addr) ;
		}
	}

	// program swins
	for ( i =0 ; i < ARTNET_MAX_PORTS ; i++) {
		if (p->data.addr.swin[i] == PROGRAM_NO_CHANGE)  {
			continue ;
		} else if (p->data.addr.swin[i] == PROGRAM_DEFAULTS)  {
			// reset to defaults
			n->ports.in[i].port_addr = _make_addr(n->state.subnet, n->ports.in[i].port_default_addr) ;
			n->ports.in[i].port_net_ctl = FALSE;

		} else if( p->data.addr.swin[i] & PROGRAM_CHANGE_MASK) {
			n->ports.in[i].port_addr = _make_addr(n->state.subnet, p->data.addr.swin[i] ) ;
			n->ports.in[i].port_net_ctl = TRUE ;
		}
	}

	// program swouts
	for ( i =0 ; i < ARTNET_MAX_PORTS ; i++) {
		if (p->data.addr.swout[i] == PROGRAM_NO_CHANGE)  {
			continue ;
		} else if (p->data.addr.swout[i] == PROGRAM_DEFAULTS)  {
			// reset to defaults
			n->ports.out[i].port_addr = _make_addr(n->state.subnet, n->ports.out[i].port_default_addr);
			n->ports.out[i].port_net_ctl = FALSE;
			n->ports.out[i].enabled = TRUE ;
		} else if( p->data.addr.swout[i] & PROGRAM_CHANGE_MASK) {
			n->ports.out[i].port_addr = _make_addr(n->state.subnet, p->data.addr.swout[i] ) ;
			n->ports.in[i].port_net_ctl = TRUE;
			n->ports.out[i].enabled = TRUE ;			
		}
	}

	// reset sequence numbers if the addresses change
	for(i=0; i< ARTNET_MAX_PORTS; i++) {
		if (addr[i] != n->ports.in[i].port_addr) 
			n->ports.in[i].seq = 0 ;
	}

	// check command
	switch (p->data.addr.command) {
		case ARTNET_PC_CANCEL:
			// fix me
			break;
		case ARTNET_PC_RESET:
			n->ports.out[0].port_status = n->ports.out[0].port_status & ~PORT_STATUS_DMX_SIP & ~PORT_STATUS_DMX_TEST & ~PORT_STATUS_DMX_TEXT ;
			// need to force a rerun of short tests here
			break ;
		case ARTNET_PC_MERGE_LTP_O:
			n->ports.out[0].merge_mode = ARTNET_MERGE_LTP ;
			n->ports.out[0].port_status = n->ports.out[0].port_status | PORT_STATUS_LPT_MODE ;
			break ;
		case ARTNET_PC_MERGE_LTP_1:
			n->ports.out[1].merge_mode = ARTNET_MERGE_LTP ;
			n->ports.out[1].port_status = n->ports.out[1].port_status | PORT_STATUS_LPT_MODE ;
			break ;
		case ARTNET_PC_MERGE_LTP_2:
			n->ports.out[2].merge_mode = ARTNET_MERGE_LTP ;
			n->ports.out[2].port_status = n->ports.out[2].port_status | PORT_STATUS_LPT_MODE ;
			break ;
		case ARTNET_PC_MERGE_LTP_3:
			n->ports.out[3].merge_mode = ARTNET_MERGE_LTP ;
			n->ports.out[3].port_status = n->ports.out[3].port_status | PORT_STATUS_LPT_MODE ;
			break ;
		case ARTNET_PC_MERGE_HTP_0:
			n->ports.out[0].merge_mode = ARTNET_MERGE_HTP ;
			n->ports.out[0].port_status = n->ports.out[0].port_status | PORT_STATUS_LPT_MODE ;
			break ;
		case ARTNET_PC_MERGE_HTP_1:
			n->ports.out[1].merge_mode = ARTNET_MERGE_HTP ;
			n->ports.out[1].port_status = n->ports.out[1].port_status | PORT_STATUS_LPT_MODE ;
			break ;
		case ARTNET_PC_MERGE_HTP_2:
			n->ports.out[2].merge_mode = ARTNET_MERGE_HTP ;
			n->ports.out[2].port_status = n->ports.out[2].port_status | PORT_STATUS_LPT_MODE ;
			break ;
		case ARTNET_PC_MERGE_HTP_3:
			n->ports.out[3].merge_mode = ARTNET_MERGE_HTP ;
			n->ports.out[3].port_status = n->ports.out[3].port_status | PORT_STATUS_LPT_MODE ;
			break ;

	}

	_build_art_poll_reply(n) ;
	
	send_poll_reply(n, TRUE) ;

}

/**
 * handle art input.
 * ArtInput packets can disable input ports.
 *
 */
void handle_input(node n, artnet_packet p) {
	int i, ports ;
	
	if(check_callback(n, p, n->callbacks.input) )
		return ;

	// servers (and raw nodes) don't respond to input packets
	if(n->state.node_type != ARTNET_NODE && n->state.node_type != ARTNET_MSRV ) 
		return ;

	ports = min( p->data.ainput.numbports, ARTNET_MAX_PORTS) ;
	for(i =0; i < ports; i++) {
		if (p->data.ainput.input[i] & PORT_DISABLE_MASK) {
			// disable
			n->ports.in[i].port_status = n->ports.in[i].port_status | PORT_STATUS_DISABLED_MASK ;
		} else {
			// enable
			n->ports.in[i].port_status = n->ports.in[i].port_status & ~PORT_STATUS_DISABLED_MASK ;
		}
	}
	
	_build_art_poll_reply(n) ;
	send_poll_reply(n, TRUE) ;	

}


/***
 * handle tod request packet
 */
void handle_tod_request(node n, artnet_packet p) {

	if(check_callback(n, p, n->callbacks.todrequest) )
		return ;

	if(n->state.node_type != ARTNET_NODE ) 
		return ;

	if(p->data.todreq.command == 0x00) {
	//	for
		// bcast tod
		send_tod_data(n) ;
	}

}

/**
 * handle tod data packet
 *
 *
 */
void handle_tod_data(node n, artnet_packet p) {

	
	if(check_callback(n, p, n->callbacks.toddata) )
		return ;

	printf("tod data\n") ;
	// no action is taken on reciept of this packet

	return;
}


/**
 * have to sort this one out. 
 *
 *
 */
void handle_ipprog(node n, artnet_packet p) {

	if(check_callback(n, p, n->callbacks.ipprog) )
		return ;

	printf("in ipprog\n") ;
}

/**
 * the main handler for an artnet packet. calls 
 * the appropriate handler function
 *
 */
int handle(node n, artnet_packet p) {

	if(check_callback(n, p, n->callbacks.recv) )
		return ;

	switch(p->type) {
		case ARTNET_POLL:
			handle_poll(n, p) ;
			break;
		case ARTNET_REPLY:
			handle_reply(n,p) ;
			break;
		case ARTNET_DMX:
			handle_dmx(n, p) ;
			break;
		case ARTNET_ADDRESS:
			handle_address(n,p) ;
			break;
		case ARTNET_INPUT:
			handle_input(n,p);
			break ;
		case ARTNET_TODREQUEST:
			handle_tod_request(n,p) ;
			break ;
		case ARTNET_TODDATA:
			handle_tod_data(n,p);
			break ;
		case ARTNET_TODCONTROL:
			printf("tod control\n") ;
			break ;
		case ARTNET_RDM:
			printf("rmd\n") ;
			break ;
		case ARTNET_VIDEOSTEUP:
			printf("vid setup\n") ;
			break ;
		case ARTNET_VIDEOPALETTE:
			printf("video palette\n") ;
			break ;
		case ARTNET_VIDEODATA:
			printf("video data\n") ;
			break ;
		case ARTNET_MACMASTER:
			printf("mac master\n") ;
			break ;
		case ARTNET_MACSLAVE:
			printf("mac slave\n") ;
			break ;
		case ARTNET_FIRMWAREMASTER:
			printf("firmware master\n") ;
			break ;
		case ARTNET_FIRMWAREREPLY:
			printf("firmware reply\n") ;
			break;
		case ARTNET_IPPROG :
			handle_ipprog(n, p) ;
			break;
		case ARTNET_IPREPLY:
			printf("ip reply\n") ;
			break ;
		case ARTNET_MEDIA:
			printf("media \n") ;
			break ;
		case ARTNET_MEDIAPATCH:
			printf("media patch\n") ;
			break ;
		case ARTNET_MEDIACONTROLREPLY:
			printf("media control reply\n") ;
			break;
		case -1:
			printf("not an ArtNet packet\n") ;
			return -1 ;
			break;
		default:
			n->state.report_code = ARTNET_RCPARSEFAIL ;
			printf("artnet but not yet implemented!, op was %hx\n", p->type) ;
	}
	return 0 ; 

}

/**
 * this gets the opcode from a packet
 */
int16_t get_type(artnet_packet p) {
	uint8_t *data ;

	if(p->length < 10) 
		return 0 ;
	if(! memcmp( &p->data, "Art-Net\0", 8)) {
		// not the best here, this needs to be tested on different arch
		data = (uint8_t *) &p->data ;

		p->type =  ( data[9] << 8) + data[8]  ;
		return p->type ;
	} else {
		return 0 ;
	}
}



/**
 * takes a subnet and an address and creates the universe address
 */
uint8_t _make_addr(uint8_t subnet, uint8_t addr) {
	return ((subnet & LOW_NIBBLE) << 4) | (addr & LOW_NIBBLE) ;
}


/**
 *
 *
 *
 */
void check_timeouts(node n, int port_id) {
	output_port_t *port ;
	time_t now ;
	time_t timeoutA, timeoutB ;
	port = &n->ports.out[port_id] ;
	time(&now) ;
	timeoutA = now - port->timeA ;
	timeoutB = now - port->timeB ;

	if( timeoutA > MERGE_TIMEOUT_SECONDS ) {
		// A is old, stop the merge
		port->ipA.s_addr = 0 ;
	} 

	if( timeoutB > MERGE_TIMEOUT_SECONDS ) {
		// B is old, stop the merge
		port->ipB.s_addr = 0 ;
	} 
}

/**
 * merge the data from two sources
 *
 */
void merge(node n, int port_id, int length) {
	int i ;
	output_port_t *port ;
	port = &n->ports.out[port_id] ;

	for(i=0;i< length;i++) {
		if(port->merge_mode == ARTNET_MERGE_HTP) 
			port->data[i] = max(port->dataA[i], port->dataB[i]) ;
		else 
			port->data[i] = min(port->dataA[i], port->dataB[i]) ;
	}
}
